"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
var ts_morph_1 = require("ts-morph");
var default_1 = require("./default");
var util_1 = require("./util");
var prettier_1 = require("prettier");
var dummyProject = new ts_morph_1.Project({ useInMemoryFileSystem: true });
var KEYWORDS_TO_SKIP = ['Omit'];
var TAG_NAMES_FOR_DESCRIPTION = ['zh', 'en'];
var internalProject = new ts_morph_1.Project({
    compilerOptions: {
        jsx: 'react',
    },
});
var propertyRegex = /(\w+)\s{0,}([?]?)\s{0,}:(.*?);?$/s;
// extract pure type
function extractFromPropertyText(text) {
    var regexResult = propertyRegex.exec(text);
    if (!regexResult) {
        return;
    }
    var name = regexResult[1];
    var isOptional = regexResult[2] === '?';
    var type = (0, util_1.toSingleLine)((0, util_1.escape)(regexResult[3]));
    return {
        name: name,
        isOptional: isOptional,
        type: type,
    };
}
// Get key-value pairs from jsDoc of Declarations
function getDeclarationTags(declaration) {
    var e_1, _a;
    var _b;
    var tags = [];
    var rawTags = ((_b = declaration.getJsDocs()[0]) === null || _b === void 0 ? void 0 : _b.getTags()) || [];
    var title;
    try {
        for (var rawTags_1 = __values(rawTags), rawTags_1_1 = rawTags_1.next(); !rawTags_1_1.done; rawTags_1_1 = rawTags_1.next()) {
            var tag = rawTags_1_1.value;
            var name = tag.getTagName();
            var value = tag.getCommentText() || '';
            if (name === 'title') {
                title = value;
            }
            tags.push({ name: name, value: value });
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (rawTags_1_1 && !rawTags_1_1.done && (_a = rawTags_1.return)) _a.call(rawTags_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return {
        title: title,
        tags: tags,
    };
}
// Get key-value pairs from jsDoc of Symbol
function getSymbolTags(sym, strictComment) {
    if (strictComment === void 0) { strictComment = false; }
    var jsDocTags = sym.compilerSymbol.getJsDocTags();
    var tags = jsDocTags.map(function (tag) {
        var _a;
        return ({
            name: tag.name,
            value: ((_a = tag.text) === null || _a === void 0 ? void 0 : _a[0].text) || '',
        });
    });
    // Try to extend property description from common comment
    if (!strictComment) {
        var _a = __read(sym.compilerSymbol.getDocumentationComment(undefined), 1), commonComment_1 = _a[0];
        if (commonComment_1 && commonComment_1.kind === 'text' && commonComment_1.text) {
            TAG_NAMES_FOR_DESCRIPTION.forEach(function (tagNameForDescription) {
                if (!tags.find(function (_a) {
                    var name = _a.name;
                    return name === tagNameForDescription;
                })) {
                    tags.push({ name: tagNameForDescription, value: commonComment_1.text });
                }
            });
        }
    }
    return tags;
}
function getSymbolByType(type) {
    return type.getAliasSymbol() || type.getSymbol();
}
function hasJSDocTitle(declaration, parsedNestedTypeSet) {
    var title = ((declaration && 'getJsDocs' in declaration && getDeclarationTags(declaration)) || {}).title;
    if (title) {
        // Type with @title in JSDoc is used to format link but not dumped as nested types onto page
        parsedNestedTypeSet.add(declaration.getType());
        return true;
    }
    return false;
}
// Check whether the type is our target that we want to continue parsing
function isTarget(type, parsedNestedTypeSet) {
    var _a;
    // Has parsed before
    if (parsedNestedTypeSet.has(type)) {
        return false;
    }
    var declaration = getDeclarationBySymbol(getSymbolByType(type));
    if (hasJSDocTitle(declaration, parsedNestedTypeSet)) {
        return false;
    }
    var defPath = (_a = declaration === null || declaration === void 0 ? void 0 : declaration.getSourceFile()) === null || _a === void 0 ? void 0 : _a.getFilePath();
    if (
    // Types from node_modules
    defPath &&
        defPath.includes('/node_modules/')) {
        return false;
    }
    return (type.isInterface() || type.isEnum() || type.isUnionOrIntersection() || isAliasDeclaration(type));
}
function isAliasDeclaration(type) {
    var _a;
    // No type.isAlias method so do alias declaration check separately
    var aliasSymbol = type.getAliasSymbol();
    if (aliasSymbol) {
        return ((_a = getDeclarationBySymbol(aliasSymbol)) === null || _a === void 0 ? void 0 : _a.getKind()) === ts_morph_1.SyntaxKind.TypeAliasDeclaration;
    }
    return false;
}
function getDeclarationBySymbol(symbol) {
    var _a;
    return (_a = symbol === null || symbol === void 0 ? void 0 : symbol.getDeclarations()) === null || _a === void 0 ? void 0 : _a[0];
}
function getDeclarationTextBySymbol(symbol) {
    var declaration = getDeclarationBySymbol(symbol);
    return (0, prettier_1.format)((declaration === null || declaration === void 0 ? void 0 : declaration.print()) || '', { parser: 'typescript' });
}
/**
 *
 * @param declaration declaration to be parsed
 * @param nestedTypeList list to store the nested types to be displayed on the doc page
 * @param parsedNestedTypes set of parsed nested types to avoid infinite loop
 * @returns
 */
function dumpNestedTypes(declaration, nestedTypeList, parsedNestedTypes) {
    if (declaration == null) {
        return;
    }
    if (hasJSDocTitle(declaration, parsedNestedTypes)) {
        return;
    }
    declaration.forEachDescendant(function (descendant) {
        var typeOfIdentifier = descendant.getType();
        var symbolOfIdentifier = getSymbolByType(typeOfIdentifier);
        var title = symbolOfIdentifier === null || symbolOfIdentifier === void 0 ? void 0 : symbolOfIdentifier.getName();
        if (
        // Only interested in type nodes that has title and matches our target check
        !ts_morph_1.Node.isTypeNode(descendant) ||
            !title ||
            KEYWORDS_TO_SKIP.includes(title) ||
            !isTarget(typeOfIdentifier, parsedNestedTypes)) {
            return;
        }
        parsedNestedTypes.add(typeOfIdentifier);
        var schema = {
            tags: [
                {
                    name: 'title',
                    value: title,
                },
            ],
            data: getDeclarationTextBySymbol(symbolOfIdentifier),
            isNestedType: true,
        };
        nestedTypeList.push({
            title: title,
            schema: schema,
        });
        if (typeOfIdentifier.isUnionOrIntersection()) {
            var method = typeOfIdentifier.isUnion() ? 'getUnionTypes' : 'getIntersectionTypes';
            // Recursively iterate subTypes
            var subTypes = typeOfIdentifier[method]() || [];
            subTypes.forEach(function (subType) {
                var subTypeSymbol = getSymbolByType(subType);
                if (subTypeSymbol && getDeclarationBySymbol(subTypeSymbol)) {
                    dumpNestedTypes(getDeclarationBySymbol(subTypeSymbol), nestedTypeList, parsedNestedTypes);
                }
            });
        }
        else if (typeOfIdentifier.isInterface()) {
            // Recursively iterate children properties
            getDeclarationBySymbol(symbolOfIdentifier)
                .getProperties()
                .forEach(function (a) {
                dumpNestedTypes(getDeclarationBySymbol(a.getSymbol()), nestedTypeList, parsedNestedTypes);
            });
        }
    });
    return nestedTypeList;
}
function getDisplayTypeWithLink(originTypeText, nestedTypeList, parsedNestedTypeSet, linkFormatter) {
    var sourceFile = dummyProject.createSourceFile('./dummy.ts', originTypeText, {
        overwrite: true,
    });
    sourceFile.transform(function (traversal) {
        var e_2, _a;
        var _b, _c;
        var node = traversal.visitChildren();
        if (ts_morph_1.ts.isIdentifier(node)) {
            var nodeTypeText = node.text;
            var _loop_1 = function (parsedNestedType) {
                var typeName = (_b = getSymbolByType(parsedNestedType)) === null || _b === void 0 ? void 0 : _b.getName();
                if (!typeName || typeName !== nodeTypeText) {
                    return "continue";
                }
                var matchedNestedType = nestedTypeList.find(function (item) { return item.title === typeName; });
                var link = void 0;
                if (matchedNestedType) {
                    // Type that is available on the current page and doesn't have JSDoc @title
                    link = linkFormatter({ typeName: typeName });
                }
                else {
                    var declaration = getDeclarationBySymbol(getSymbolByType(parsedNestedType));
                    var definitionPath = (_c = declaration === null || declaration === void 0 ? void 0 : declaration.getSourceFile()) === null || _c === void 0 ? void 0 : _c.getFilePath();
                    var title = (((declaration === null || declaration === void 0 ? void 0 : declaration.getJsDocs) && getDeclarationTags(declaration)) || {}).title;
                    // Has @title in JSDoc
                    if (title) {
                        link = linkFormatter({ typeName: typeName, jsDocTitle: title, fullPath: definitionPath });
                    }
                }
                // Only convert to link when link is available
                if (link) {
                    return { value: ts_morph_1.ts.factory.createIdentifier("[".concat(typeName, "](").concat(link, ")")) };
                }
            };
            try {
                // Check identifiers in the type against parsed custom type definitions and replace with link info if necessary
                for (var parsedNestedTypeSet_1 = __values(parsedNestedTypeSet), parsedNestedTypeSet_1_1 = parsedNestedTypeSet_1.next(); !parsedNestedTypeSet_1_1.done; parsedNestedTypeSet_1_1 = parsedNestedTypeSet_1.next()) {
                    var parsedNestedType = parsedNestedTypeSet_1_1.value;
                    var state_1 = _loop_1(parsedNestedType);
                    if (typeof state_1 === "object")
                        return state_1.value;
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (parsedNestedTypeSet_1_1 && !parsedNestedTypeSet_1_1.done && (_a = parsedNestedTypeSet_1.return)) _a.call(parsedNestedTypeSet_1);
                }
                finally { if (e_2) throw e_2.error; }
            }
        }
        return node;
    });
    return sourceFile.getText();
}
// Get Json schema of interface's property
function getPropertySchema(sym, defaultT, strictComment, nestedTypeList, parsedNestedTypeSet, linkFormatter) {
    if (strictComment === void 0) { strictComment = false; }
    var name = sym.getName();
    var declaration = sym.getDeclarations()[0];
    var typeText = declaration.getText();
    var extract = extractFromPropertyText(typeText);
    if (!extract) {
        return null;
    }
    var tags = getSymbolTags(sym, strictComment);
    if (tags.find(function (_a) {
        var name = _a.name;
        return name && TAG_NAMES_FOR_DESCRIPTION.indexOf(name) > -1;
    })) {
        var typeWithLink = extract.type;
        // Deeply analyze nested types
        dumpNestedTypes(declaration, nestedTypeList, parsedNestedTypeSet);
        typeWithLink = getDisplayTypeWithLink(typeWithLink, nestedTypeList, parsedNestedTypeSet, linkFormatter);
        return {
            name: name,
            type: typeWithLink,
            isOptional: extract.isOptional,
            tags: tags,
        };
    }
    return defaultT[name]
        ? __assign({ name: name, isOptional: extract.isOptional }, defaultT[name]) : null;
}
// Get Json schema of Function
function getFunctionSchema(declaration, strictComment, nestedTypeList, parsedNestedTypeSet, linkFormatter) {
    if (strictComment === void 0) { strictComment = false; }
    return {
        params: declaration.getParameters().map(function (para) {
            var _a, _b;
            var tags = getSymbolTags(para.getSymbol(), strictComment);
            var typeWithLink = para
                .getType()
                .getText()
                .replace(/import\([^)]+\)\./g, '');
            // Deeply analyze nested types
            dumpNestedTypes(para, nestedTypeList, parsedNestedTypeSet);
            typeWithLink = getDisplayTypeWithLink(typeWithLink, nestedTypeList, parsedNestedTypeSet, linkFormatter);
            return {
                tags: tags,
                name: para.getName(),
                type: typeWithLink,
                isOptional: para.isOptional(),
                initializerText: ((_a = para.getInitializer()) === null || _a === void 0 ? void 0 : _a.getText()) ||
                    ((_b = tags.find(function (_a) {
                        var name = _a.name;
                        return name === 'default' || name === 'defaultValue';
                    })) === null || _b === void 0 ? void 0 : _b.value) ||
                    null,
            };
        }),
        returns: declaration.getReturnType().getText(),
    };
}
function generateSchema(sourceFile, typeChecker, config) {
    var interfaces = (sourceFile === null || sourceFile === void 0 ? void 0 : sourceFile.getInterfaces()) || [];
    var typeAliases = (sourceFile === null || sourceFile === void 0 ? void 0 : sourceFile.getTypeAliases()) || [];
    var functions = (sourceFile === null || sourceFile === void 0 ? void 0 : sourceFile.getFunctions()) || [];
    var defaultT = (config === null || config === void 0 ? void 0 : config.defaultTypeMap) || default_1.defaultTypeMap;
    var strictComment = !!(config === null || config === void 0 ? void 0 : config.strictComment);
    var propertySorter = config === null || config === void 0 ? void 0 : config.propertySorter;
    var linkFormatter = (config === null || config === void 0 ? void 0 : config.linkFormatter) || default_1.defaultLinkFormatter;
    var schemaMap = {};
    var schemaList = [];
    var nestedTypeList = [];
    var parsedNestedTypeSet = new Set();
    __spreadArray(__spreadArray(__spreadArray([], __read(interfaces), false), __read(typeAliases), false), __read(functions), false).sort(function (declarationA, declarationB) {
        return declarationA.getStartLineNumber() - declarationB.getStartLineNumber();
    })
        .forEach(function (declaration) {
        var _a, _b;
        var _c = getDeclarationTags(declaration), title = _c.title, tags = _c.tags;
        var dType = declaration.getKindName();
        if (!title) {
            return;
        }
        var schema;
        var typeNode = dType === 'FunctionDeclaration'
            ? declaration
            : dType === 'TypeAliasDeclaration'
                ? declaration.getTypeNode()
                : null;
        // Function declaration
        if (typeNode &&
            ['FunctionDeclaration', 'FunctionType'].indexOf(typeNode.getKindName()) > -1) {
            schema = __assign({ tags: tags }, getFunctionSchema(typeNode, strictComment, nestedTypeList, parsedNestedTypeSet, linkFormatter));
        }
        // Interface declaration forbid extends
        else if (dType === 'InterfaceDeclaration' &&
            !!tags.find(function (_a) {
                var name = _a.name;
                return name === 'notExtends';
            })) {
            var data_1 = [];
            declaration.getProperties().forEach(function (a) {
                var schema = getPropertySchema(a.getSymbol(), defaultT, strictComment, nestedTypeList, parsedNestedTypeSet, linkFormatter);
                schema && data_1.push(schema);
            });
            schema = { tags: tags, data: data_1 };
        }
        else {
            var data_2 = [];
            typeChecker.getPropertiesOfType(declaration.getType()).forEach(function (a) {
                var schema = getPropertySchema(a, defaultT, strictComment, nestedTypeList, parsedNestedTypeSet, linkFormatter);
                schema && data_2.push(schema);
            });
            schema = { tags: tags, data: data_2 };
        }
        if (typeof propertySorter === 'function') {
            (_a = schema.data) === null || _a === void 0 ? void 0 : _a.sort(propertySorter);
            (_b = schema.params) === null || _b === void 0 ? void 0 : _b.sort(propertySorter);
        }
        schemaList.push({ title: title, schema: schema });
        schemaMap[title] = schema;
    });
    var list = schemaList;
    var map = schemaMap;
    if (nestedTypeList.length > 0) {
        list = __spreadArray(__spreadArray([], __read(schemaList), false), __read(nestedTypeList), false);
        map = __assign(__assign({}, schemaMap), nestedTypeList.reduce(function (result, item) {
            result[item.title] = item.schema;
            return result;
        }, {}));
    }
    return (config === null || config === void 0 ? void 0 : config.strictDeclarationOrder) ? list : map;
}
function generate(file, config) {
    var project = (config === null || config === void 0 ? void 0 : config.project) || internalProject;
    if (config === null || config === void 0 ? void 0 : config.sourceFilesPaths) {
        project.addSourceFilesAtPaths(config === null || config === void 0 ? void 0 : config.sourceFilesPaths);
    }
    var typeChecker = project.getTypeChecker();
    var sourceFile = project.getSourceFile(file);
    if (!sourceFile) {
        return;
    }
    return generateSchema(sourceFile, typeChecker, config);
}
exports.default = generate;
